/* TimelineDayDisplay.java
 * Author: Jonathan Bernard - jonathan.bernard@gemalto.com
 */

package jdbernard.timestamper;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 *
 * @author jbernard
 */
public class TimelineDayDisplay extends JComponent implements MouseListener {

    private class MarkerDisplayEntry {
        public Timeline.TimelineMarker marker;
        public float relY;
        public float relHeight;
        public Rectangle2D markBounds;
        public Rectangle2D notesBounds;
        public Rectangle bounds;
    }

    private Timeline timeline;
    private ArrayList<MarkerDisplayEntry> markerEntries;
    private Calendar day = Calendar.getInstance();
    private Calendar nextDay = Calendar.getInstance();
    private Timeline.TimelineMarker currentMarker;
    private ArrayList<ChangeListener> changeListeners = new ArrayList<ChangeListener>();

    private Font markFont;// = getFont().deriveFont(Font.BOLD);
    private Font notesFont;// = getFont();

    private Color evenTrans = new Color(0.75f, 0.75f, 0.75f, 0.4f);
    private Color evenOpaque = new Color(0.75f, 0.75f, 0.75f, 1f);
    private Color oddTrans = new Color(0.5f, 0.5f, 0.5f, 0.4f);
    private Color oddOpaque = new Color(0.5f, 0.5f, 0.5f, 1f);
    private Color selectedTrans = new Color(0.5f, 0.75f, 0.5f, 0.4f);
    private Color selectedOpaque = new Color(0.5f, 0.75f, 0.5f, 1f);
    private Color fontColor = new Color(0.1f, 0.1f, 0.1f, 1f);

    private long oneDayInMilli = (24 * 60 * 60 * 1000);

    public TimelineDayDisplay() {
        super();
        addMouseListener(this);
    }

    public TimelineDayDisplay(Timeline timeline, Calendar day) {
        this();
        this.timeline = timeline;
        this.day = day;
        updateMarkers(getGraphics());
    }

    public void setTimeline(Timeline t) {
        timeline = t;
        updateMarkers(getGraphics());
    }

    public Timeline getTimeline() {
        return timeline;
    }

    public void setDay(Calendar d) {
        day.setTime(d.getTime());
        day.set(Calendar.HOUR_OF_DAY, 0);
        day.set(Calendar.MINUTE, 0);
        day.set(Calendar.SECOND, 0);

        nextDay.setTime(day.getTime());
        nextDay.add(Calendar.DAY_OF_YEAR, 1);
        updateMarkers(getGraphics());
    }

    public Calendar getDay() {
        return day;
    }

    public void setMarkFont(Font f) {
        markFont = f;
    }

    public Font getMarkFont() {
        return markFont;
    }

    public void setNotesFont(Font f) {
        notesFont = f;
    }

    public Font getNotesFont() {
        return notesFont;
    }

    public void setFontColor(Color f) {
        fontColor = new Color(f.getRGB());
    }

    public Color getFontColor() {
        return fontColor;
    }
    public void setEvenColor(Color c) {
        evenOpaque = new Color(c.getRGB());
        evenTrans = new Color((float) c.getRed() / 255f,
                (float) c.getGreen() / 255f,
                (float) c.getBlue() / 255f, 0.4f);
    }

    public Color getEvenColor() {
        return evenOpaque;
    }

    public void setOddColor(Color c) {
        oddOpaque = new Color(c.getRGB());
        oddTrans = new Color((float) c.getRed() / 255f,
                (float) c.getGreen() / 255f,
                (float) c.getBlue() / 255f, 0.4f);
    }

    public Color getOddColor() {
        return oddOpaque;
    }

    public void setSelectedColor(Color c) {
        selectedOpaque = new Color(c.getRGB());
        selectedTrans = new Color((float) c.getRed() / 255f,
                (float) c.getGreen() / 255f,
                (float) c.getBlue() / 255f, 0.4f);
    }

    public Color getSelectedColor() {
        return selectedOpaque;
    }

    public Timeline.TimelineMarker getSelectedTimelineMarker() {
        return currentMarker;
    }

    public void addMarker(Date timestamp, String mark, String notes) {
        timeline.addMarker(timestamp, mark, notes);
        updateMarkers(getGraphics());
    }

    public void deleteSelectedMarker() {
        timeline.removeMarker(currentMarker);
        updateMarkers(getGraphics());
    }

    public void updateSelectedMarker(String notes) {
        currentMarker.setNotes(notes);
        updateMarkers(getGraphics());
    }

    private void updateMarkers(Graphics g) {

        markerEntries = new ArrayList<MarkerDisplayEntry>();

        if (markFont == null) markFont = getFont().deriveFont(Font.BOLD);
        if (notesFont == null) notesFont = getFont();

        // get all relevant markers
        Timeline.TimelineMarker tm = timeline.getLastMarker(day.getTime());
        Iterator<Timeline.TimelineMarker> itr = timeline.iterator();

        while (!itr.next().equals(tm));

        ArrayList<Timeline.TimelineMarker> markers = new ArrayList<Timeline.TimelineMarker>();
        while (nextDay.getTime().after(tm.getTimestamp())) {
            markers.add(tm);
            if (itr.hasNext()) tm = itr.next();
            else break;
        }

        markers.add(tm);

        for (int i = 0; i < markers.size() - 1; i++) {
            MarkerDisplayEntry markerEntry = new MarkerDisplayEntry();

            markerEntry.marker = markers.get(i);

            // set string bounds
            markerEntry.markBounds = getFontMetrics(markFont)
                    .getStringBounds(markers.get(i).getMark(), g);
            markerEntry.notesBounds = getFontMetrics(notesFont)
                    .getStringBounds(markers.get(i).getNotes(), g);

            // calculate upper bound
            if (i == 0) {
                markerEntry.relY = 0;
            } else {
                // calculate time in ms since the beginning of the day
                markerEntry.relY = markers.get(i).getTimestamp().getTime()
                        - day.getTimeInMillis();
                // calculate percentage of total time in day
                markerEntry.relY /= oneDayInMilli;
            }

            // calculate lower bound
            if (i == 0)
                markerEntry.relHeight =
                        markers.get(i + 1).getTimestamp().getTime()
                        - day.getTimeInMillis();
            else if (i == markers.size() - 2)
                markerEntry.relHeight = nextDay.getTimeInMillis()
                        - markers.get(i).getTimestamp().getTime();
            else
                markerEntry.relHeight =
                        markers.get(i + 1).getTimestamp().getTime()
                        - markers.get(i).getTimestamp().getTime();
            markerEntry.relHeight /= oneDayInMilli;

            markerEntries.add(markerEntry);
        }
    }

    @Override
    public void paintComponent(Graphics g) {
        removeAll();

        if (markerEntries == null) updateMarkers(g);

        Insets insets = this.getInsets();
        Rectangle bounds = this.getBounds();
        Rectangle canvasBounds = new Rectangle(insets.left, insets.top,
                bounds.width - insets.left - insets.right - 1,
                bounds.height - insets.top - insets.bottom - 1);
        double hourHeight = canvasBounds.getHeight() / 24.0;

        Graphics2D g2d = (Graphics2D) g;
        Rectangle2D stringBounds = getFontMetrics(getFont()).getStringBounds("00:00 ", g);

        // draw hour lines
        for (int i = 1; i < 24; i++) {
            int lineY = (int) Math.round(hourHeight * (double) i);
            g.drawLine(canvasBounds.x + (int) stringBounds.getWidth(), canvasBounds.y + lineY,
                canvasBounds.x + canvasBounds.width,
                canvasBounds.y + lineY);

            //draw hours labels
            if ((hourHeight > (stringBounds.getHeight()) || ((i % 2) == 0))) {
                g.drawString(String.format("%1$02d:00", i),
                        canvasBounds.x + 2, lineY + (int) (stringBounds.getHeight() / 2));
            }
        }

        for (int i = 0; i < markerEntries.size(); i++) {

            MarkerDisplayEntry curEntry = markerEntries.get(i);

            Rectangle2D markBounds;
            Rectangle2D notesBounds;

            boolean selected = curEntry.marker.equals(currentMarker);

            // if i == 0, this is the default
            curEntry.bounds = new Rectangle();
            curEntry.bounds.y = 3;
            curEntry.bounds.x = canvasBounds.x + (int) stringBounds.getWidth() + 5;
            curEntry.bounds.height = 1;
            curEntry.bounds.width = canvasBounds.width - (int) stringBounds.getWidth() - 8;

            double relTime;

            // calculate upper bound
            curEntry.bounds.y = (int) Math.round(curEntry.relY
                    * canvasBounds.getHeight());

            if (i == 0) curEntry.bounds.y += 3;

            // calculate lower bound
            curEntry.bounds.height = (int) Math.round(curEntry.relHeight
                    * canvasBounds.getHeight());

            if (i ==0) curEntry.bounds.height -= 6;
            else curEntry.bounds.height -= 3;

            // draw box
            if (selected) g.setColor(selectedTrans);
            else g.setColor((i % 2 == 0 ? evenTrans : oddTrans));
            g.fillRect(curEntry.bounds.x, curEntry.bounds.y, curEntry.bounds.width, curEntry.bounds.height);

            if (selected) g.setColor(selectedOpaque);
            else g2d.setColor((i % 2 == 0 ? evenOpaque : oddOpaque));
            g2d.setStroke(new BasicStroke(3f));
            g2d.drawRect(curEntry.bounds.x, curEntry.bounds.y, curEntry.bounds.width, curEntry.bounds.height);

            // draw timestamp name
            markBounds = (Rectangle2D) curEntry.markBounds.clone();
            markBounds.setRect(curEntry.bounds.x + 3,
                    curEntry.bounds.y + stringBounds.getHeight(),
                    markBounds.getWidth(), markBounds.getHeight());

            g.setColor(fontColor);
            g.setFont(markFont);
            g.drawString(curEntry.marker.getMark(),
                    (int) markBounds.getX(), (int) markBounds.getY());

            // draw notes
            notesBounds = (Rectangle2D) curEntry.notesBounds.clone();
            notesBounds.setRect(curEntry.bounds.x + 6,
                    curEntry.bounds.y + stringBounds.getHeight() + markBounds.getHeight(),
                    notesBounds.getWidth(), notesBounds.getHeight());

            if (curEntry.bounds.contains(notesBounds)) {
                g.setFont(notesFont);
                g.drawString(curEntry.marker.getNotes(),
                        (int) notesBounds.getX(), (int) notesBounds.getY());
            }
        }

        g.setColor(Color.BLACK);
        g.drawRect(canvasBounds.x, canvasBounds.y, canvasBounds.width, canvasBounds.height);
    }

    public void addChangeListener(ChangeListener cl) {
        changeListeners.add(cl);
    }

    public boolean removeChangeListener(ChangeListener cl) {
        return changeListeners.remove(cl);
    }

    private void fireChangeEvent() {
        for (ChangeListener cl : changeListeners)
            cl.stateChanged(new ChangeEvent(this));
    }

    public void mouseClicked(MouseEvent e) {
        Point topLeft = getLocationOnScreen();
        for (MarkerDisplayEntry markerEntry : markerEntries) {
            Rectangle absBounds = new Rectangle(markerEntry.bounds);
            absBounds.translate(topLeft.x, topLeft.y);
            if (absBounds.contains(e.getLocationOnScreen())) {
                currentMarker = markerEntry.marker;
                repaint();
            }
        }
        fireChangeEvent();
    }

    public void mousePressed(MouseEvent e) {
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }
}
